/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef COMPONENT_BASED_APP_CONTROL_COMPONENT_BASED_APP_CONTROL_H_
#define COMPONENT_BASED_APP_CONTROL_COMPONENT_BASED_APP_CONTROL_H_

#include <app_control.h>

#include <list>
#include <map>
#include <memory>
#include <string>

#include "component_based/uri/component_based_uri.h"

#ifndef EXPORT_API
#define EXPORT_API __attribute__((visibility("default")))
#endif

namespace component_based {

class EXPORT_API AppControl {
 public:
  enum class Error : int {
    NONE = APP_CONTROL_ERROR_NONE,
    INVALID_PARAMETER = APP_CONTROL_ERROR_INVALID_PARAMETER,
    OUT_OF_MEMORY = APP_CONTROL_ERROR_OUT_OF_MEMORY,
    APP_NOT_FOUND = APP_CONTROL_ERROR_APP_NOT_FOUND,
    KEY_NOT_FOUND = APP_CONTROL_ERROR_KEY_NOT_FOUND,
    KEY_REJECTED = APP_CONTROL_ERROR_KEY_REJECTED,
    INVALID_DATA_TYPE = APP_CONTROL_ERROR_INVALID_DATA_TYPE,
    LAUNCH_REJECTED = APP_CONTROL_ERROR_LAUNCH_REJECTED,
    PERMISSION_DENIED = APP_CONTROL_ERROR_PERMISSION_DENIED,
    LAUNCH_FAILED = APP_CONTROL_ERROR_LAUNCH_FAILED,
    TIMED_OUT = APP_CONTROL_ERROR_TIMED_OUT,
    IO_ERROR = APP_CONTROL_ERROR_IO_ERROR,
  };

  enum class Result : int {
    APP_STARTED = APP_CONTROL_RESULT_APP_STARTED,
    SUCCEEDED = APP_CONTROL_RESULT_SUCCEEDED,
    FAILED = APP_CONTROL_RESULT_FAILED,
    CANCELED = APP_CONTROL_RESULT_CANCELED,
  };

  enum class LaunchMode : int {
    SINGLE = APP_CONTROL_LAUNCH_MODE_SINGLE,
    GROUP = APP_CONTROL_LAUNCH_MODE_GROUP,
  };

  class IEventListener {
   public:
    virtual void OnResultReceived(AppControl request,
                                  AppControl::Error error) = 0;
    virtual void OnReplyReceived(AppControl request, AppControl reply,
                                 AppControl::Result result) = 0;
  };

  class Builder {
   public:
    virtual ~Builder();
    Builder& SetAppID(std::string appid);
    Builder& SetOperation(std::string op);
    Builder& SetUri(Uri uri);
    Builder& SetMime(std::string mime);
    Builder& AddExtra(std::string key, std::string val);
    Builder& SetLaunchMode(AppControl::LaunchMode mode);
    Builder& SetComponentID(std::string comp_id);
    AppControl Build();

   private:
    app_control_h BuildNative();

   private:
    std::string appid_;
    std::string op_;
    Uri uri_;
    std::string mime_;
    std::map<std::string, std::string> extra_;
    AppControl::LaunchMode mode_ = AppControl::LaunchMode::SINGLE;
    std::string comp_id_;
  };

  explicit AppControl(app_control_h control);
  virtual ~AppControl();

  AppControl(const AppControl& control);
  AppControl& operator=(const AppControl& control);

  AppControl(AppControl&& control);
  AppControl& operator=(AppControl&& control);

  std::string GetAppID();
  std::string GetOperation();
  Uri GetUri();
  std::string GetMime();
  std::string GetExtra(std::string key);
  LaunchMode GetLaunchMode();
  std::string GetComponentID();

  app_control_h GetHandle() const;

 private:
  class Impl;
  std::unique_ptr<Impl> impl_;
};

}  // namespace component_based

#endif  // COMPONENT_BASED_APP_CONTROL_COMPONENT_BASED_APP_CONTROL_H_
