/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef COMPONENT_BASED_BASE_COMPONENT_H_
#define COMPONENT_BASED_BASE_COMPONENT_H_

#include <bundle_cpp.h>

#include <memory>
#include <string>

#include "component_based/app_control/component_based_app_control.h"
#include "component_based/base/device_orientation.h"
#include "component_based/base/low_battery.h"
#include "component_based/base/low_memory.h"
#include "component_based/base/suspended_state.h"

#ifndef EXPORT_API
#define EXPORT_API __attribute__((visibility("default")))
#endif

namespace component_based {

class EXPORT_API Component {
 public:
  enum class State {
    None,
    Created,
    Started,
    Running,
    Paused,
    Stopped,
    Dying,
  };

  enum class Type {
    Frame,
    Service,
  };

  class Factory {
   public:
    virtual std::unique_ptr<Component> Create(std::string comp_id,
                                              std::string inst_id) = 0;
  };

  Component(Type type, std::string comp_id, std::string inst_id);
  virtual ~Component();

  virtual bool OnBaseCreate();
  virtual void OnBaseDestroy();
  virtual void OnBaseStart(AppControl control, bool restarted);
  virtual void OnBaseResume();
  virtual void OnBasePause();
  virtual void OnBaseStop();
  virtual void OnBaseRestoreContent(tizen_base::Bundle content);
  virtual void OnBaseSaveContent(tizen_base::Bundle& content);
  virtual void OnBaseAction(std::string action, AppControl app_control);
  virtual void OnBaseDeviceOrientationChanged(
      DeviceOrientation::Orientation orientation);
  virtual void OnBaseLanguageChanged(std::string language);
  virtual void OnBaseRegionFormatChanged(std::string region);
  virtual void OnBaseLowBattery(LowBattery::Status status);
  virtual void OnBaseLowMemory(LowMemory::Status status);
  virtual void OnBaseSuspendedStateChanged(SuspendedState::State state);

  std::string GetComponentID();
  std::string GetInstanceID();
  void Finish();
  Type GetType();
  void RegisterAction(std::string app_control_name);
  bool DeregisterAction(std::string app_control_name);
  void SendAsync(AppControl control, AppControl::IEventListener* ev);
  AppControl Send(AppControl control);

 protected:
  State GetState();
  void SetState(State state);
  tizen_base::Bundle GetContent();
  void SetContent(tizen_base::Bundle content);

 private:
  class Impl;
  std::unique_ptr<Impl> impl_;
};

}  // namespace component_based

#endif  // COMPONENT_BASED_BASE_COMPONENT_H_
