/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <aul.h>
#include <aul_app_group.h>
#include <app_control_internal.h>
#include <bundle_internal.h>
#include <glib.h>

#include <algorithm>
#include <memory>
#include <string>

#include "component_based/common/exception.h"
#include "component_based/base/component_manager_internal.h"
#include "component_based/base/dlog_internal.h"

namespace component_based {
namespace internal {

void ComponentManager::RegisterComponent(const std::string& comp_id,
    std::unique_ptr<Component::Factory> factory) {
  auto i = fac_map_.find(comp_id);
  if (i == fac_map_.end()) {
    appcore_multiwindow_base_class cls =
        appcore_multiwindow_base_class_get_default();
    cls.id = strdup(comp_id.c_str());
    if (cls.id == nullptr)
      throw std::bad_alloc();
    cls.data = this;
    cls.create = OnCreateCB;
    cls.terminate = OnTerminateCB;
    cls.pause = OnPauseCB;
    cls.resume = OnResumeCB;
    appcore_multiwindow_base_class_add(cls);
  }

  fac_map_[comp_id] = std::move(factory);
}

void ComponentManager::AddEventListener(const std::string& inst_id,
    ComponentManager::IEventListener* ev) {
  ev_map_[inst_id] = ev;
}

void ComponentManager::Start(const std::string& comp_id,
                             const std::string& inst_id,
                             bundle* b) {
  bool restarted = true;
  auto inst_iter = inst_map_.find(inst_id);
  if (inst_iter == inst_map_.end()) {
    auto fac_iter = fac_map_.find(comp_id);
    if (fac_iter == fac_map_.end()) {
      LOGE("Failed to find factory map. %s", comp_id.c_str());
      return;
    }

    const char *val = bundle_get_val(b, AUL_K_NEW_INSTANCE);
    std::string new_inst(val ? val : "");
    if (new_inst == "true")
      multi_inst_list_.push_back(inst_id);

    std::unique_ptr<Component> component = fac_map_[comp_id]->Create(comp_id,
        inst_id);
    if (component == nullptr || component.get() == nullptr) {
      LOGE("Fail to create component");
      return;
    }
    inst_map_[inst_id] = std::move(component);
    appcore_multiwindow_base_instance_run(comp_id.c_str(), inst_id.c_str(),
        nullptr);
    restarted = false;
  }

  auto ev_iter = ev_map_.find(inst_id);
  if (ev_iter == ev_map_.end())
    return;

  app_control_h handle = nullptr;
  app_control_create_event(b, &handle);
  try {
    AppControl control(handle);
    ev_map_[inst_id]->OnStart(control, restarted);
  } catch(Exception& ex) {
    LOGE("Exception occurred (%s)", ex.what());
  }
}

void ComponentManager::Resume(const std::string& inst_id) {
  Ecore_Wl2_Window* wl_win = GetWl2Window(inst_id);
  if (wl_win == nullptr)
    return;
  ecore_wl2_window_activate(wl_win);
}

void ComponentManager::Pause(const std::string& inst_id) {
  Ecore_Wl2_Window* wl_win = GetWl2Window(inst_id);
  if (wl_win == nullptr)
    return;
  ecore_wl2_window_iconified_set(wl_win, EINA_TRUE);
}

void ComponentManager::Stop(const std::string& inst_id) {
  auto ev_iter = ev_map_.find(inst_id);
  if (ev_iter == ev_map_.end())
    return;

  ev_map_[inst_id]->OnPause();
  ev_map_[inst_id]->OnStop();
}

bool ComponentManager::IsResumed(const std::string& inst_id) {
  appcore_multiwindow_base_instance_h context =
    appcore_multiwindow_base_instance_find(inst_id.c_str());
  if (!context) {
    LOGE("Failed to find instance(%s)", inst_id.c_str());
    return false;
  }

  return appcore_multiwindow_base_instance_is_resumed(context);
}

void ComponentManager::Exit(const std::string& inst_id) {
  appcore_multiwindow_base_instance_h context =
      appcore_multiwindow_base_instance_find(inst_id.c_str());
  if (!context) {
    LOGE("Failed to find instance(%s)", inst_id.c_str());
    return;
  }

  appcore_multiwindow_base_instance_exit(context);
}

void ComponentManager::ExitAtIdle(const std::string& inst_id) {
  g_idle_add([](gpointer data) -> gboolean {
        char* inst_id = static_cast<char*>(data);
        ComponentManager::GetInst().Exit(inst_id);
        free(inst_id);
        return G_SOURCE_REMOVE;
      }, strdup(inst_id.c_str()));
}

void ComponentManager::ExitAll() {
  if (inst_map_.empty())
    return;

  auto iter = inst_map_.begin();
  while (iter != inst_map_.end()) {
    auto i = iter;
    LOGD("Exit instance(%s)", i->first.c_str());
    iter++;
    Exit(i->first);
  }
}

bool ComponentManager::Bind(const std::string& inst_id,
                            std::unique_ptr<IWindow> win) {
  appcore_multiwindow_base_instance_h context =
      appcore_multiwindow_base_instance_find(inst_id.c_str());
  if (!context) {
    LOGE("Failed to find instance(%s)", inst_id.c_str());
    return false;
  }

  win_map_[inst_id] = std::move(win);
  Ecore_Wl2_Window* wl_win = GetWl2Window(inst_id);
  if (!wl_win) {
    LOGE("Failed to get wayland window(%s)", inst_id.c_str());
    return false;
  }

  appcore_multiwindow_base_window_bind(context, wl_win);
  LOGI("Bind instance(%s)", inst_id.c_str());
  return true;
}

void ComponentManager::Unbind(const std::string& inst_id) {
  auto win_iter = win_map_.find(inst_id);
  if (win_iter == win_map_.end()) {
    return;
  }

  appcore_multiwindow_base_instance_h context =
      appcore_multiwindow_base_instance_find(inst_id.c_str());
  if (context) {
    appcore_multiwindow_base_window_unbind(context);
  }
  win_map_.erase(inst_id);
  LOGI("Unbind instance(%s)", inst_id.c_str());
}

ComponentManager::IEventListener* ComponentManager::GetEventListener(
    const std::string& inst_id) {
  auto ev_iter = ev_map_.find(inst_id);
  if (ev_iter == ev_map_.end())
    return nullptr;

  return ev_map_[inst_id];
}

std::string ComponentManager::GetInstanceID(int win_id) {
  for (auto i = win_map_.begin(); i != win_map_.end(); i++) {
    auto& win = *i->second;
    if (win.GetResID() == win_id)
      return i->first;
  }

  return std::string("");
}

const IWindow* ComponentManager::GetWindow(const std::string& inst_id) {
  auto win_iter = win_map_.find(inst_id);
  if (win_iter == win_map_.end())
    return nullptr;

  return &*win_map_[inst_id];
}

void ComponentManager::AddGroup(const std::string& inst_id, int surf_id) {
  auto group_iter = group_map_.find(inst_id);
  if (group_iter != group_map_.end())
    return;

  LOGD("Group attach. instance(%s), surf(%d)", inst_id.c_str(), surf_id);

  aul_app_group_set_window_v2(inst_id.c_str(), surf_id);
  group_map_[inst_id] = surf_id;
}

void ComponentManager::RemoveGroup(const std::string& inst_id) {
  auto group_iter = group_map_.find(inst_id);
  if (group_iter == group_map_.end())
    return;

  LOGD("Group lower. instance(%s)", inst_id.c_str());
  bool exit = false;
  aul_app_group_lower_v2(inst_id.c_str(), &exit);
  if (exit) {
    Exit(inst_id);
  }

  group_map_.erase(inst_id);
}

bool ComponentManager::IsEmpty() {
  return inst_map_.empty();
}

void ComponentManager::ResumeAll() {
  if (inst_map_.empty())
    return;

  for (auto i = inst_map_.begin(); i != inst_map_.end(); i++) {
    LOGD("Resume instance(%s)", i->first.c_str());
    Resume(i->first);
  }
}

void ComponentManager::PauseAll() {
  if (inst_map_.empty())
    return;

  for (auto i = inst_map_.begin(); i != inst_map_.end(); i++) {
    LOGD("Pause instance(%s)", i->first.c_str());
    Pause(i->first);
  }
}

bool ComponentManager::IsSingleInstance(const std::string& inst_id) {
  auto it = std::find(multi_inst_list_.begin(), multi_inst_list_.end(),
      inst_id);
  if (it == multi_inst_list_.end())
    return true;

  return false;
}

Ecore_Wl2_Window* ComponentManager::GetWl2Window(const std::string& inst_id) {
  auto win_iter = win_map_.find(inst_id);
  if (win_iter == win_map_.end())
    return nullptr;

  int win_id = win_map_[inst_id]->GetResID();
  Ecore_Wl2_Display* display = ecore_wl2_connected_display_get(NULL);
  Ecore_Wl2_Window* win = ecore_wl2_display_window_find(display, win_id);
  return win;
}

void ComponentManager::Remove(const std::string& inst_id) {
  multi_inst_list_.remove(inst_id);
  RemoveGroup(inst_id);
  ev_map_.erase(inst_id);
  inst_map_.erase(inst_id);
  win_map_.erase(inst_id);
}

void ComponentManager::OnCreateCB(appcore_multiwindow_base_instance_h context,
                                  void* data) {
  std::string inst_id = appcore_multiwindow_base_instance_get_id(context);
  auto& mgr = ComponentManager::GetInst();
  ComponentManager::IEventListener* ev = mgr.GetEventListener(inst_id);
  if (!ev->OnCreate()) {
    mgr.Exit(inst_id);
  }
}

void ComponentManager::OnTerminateCB(
    appcore_multiwindow_base_instance_h context, void* data) {
  std::string inst_id = appcore_multiwindow_base_instance_get_id(context);
  auto& mgr = ComponentManager::GetInst();
  ComponentManager::IEventListener* ev = mgr.GetEventListener(inst_id);
  ev->OnPause();
  ev->OnStop();
  ev->OnDestroy();
  mgr.Remove(inst_id);

  if (appcore_multiwindow_base_instance_get_cnt() == 1) {
    LOGD("Exit");
    appcore_multiwindow_base_exit();
  }
}

void ComponentManager::OnPauseCB(appcore_multiwindow_base_instance_h context,
                                 void* data) {
  std::string inst_id = appcore_multiwindow_base_instance_get_id(context);
  auto& mgr = ComponentManager::GetInst();
  ComponentManager::IEventListener* ev = mgr.GetEventListener(inst_id);
  ev->OnPause();
}

void ComponentManager::OnResumeCB(appcore_multiwindow_base_instance_h context,
                                  void* data) {
  std::string inst_id = appcore_multiwindow_base_instance_get_id(context);
  auto& mgr = ComponentManager::GetInst();
  ComponentManager::IEventListener* ev = mgr.GetEventListener(inst_id);
  ev->OnResume();
}

}  // namespace internal
}  // namespace component_based
