/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef COMPONENT_BASED_BASE_COMPONENT_MANAGER_INTERNAL_H_
#define COMPONENT_BASED_BASE_COMPONENT_MANAGER_INTERNAL_H_

#include <appcore_multiwindow_base.h>

#include <list>
#include <map>
#include <memory>
#include <string>

#include "component_based/base/component.h"
#include "component_based/base/window_interface.h"

namespace component_based {
namespace internal {

class ComponentManager {
 private:
  ComponentManager() = default;
  ~ComponentManager() = default;

 public:
  class IEventListener {
   public:
    virtual bool OnCreate() = 0;
    virtual void OnDestroy() = 0;
    virtual void OnStart(AppControl app_control, bool restarted) = 0;
    virtual void OnPause() = 0;
    virtual void OnResume() = 0;
    virtual void OnStop() = 0;
  };

 public:
  static ComponentManager& GetInst() {
    static ComponentManager inst;
    return inst;
  }

  void AddEventListener(const std::string& inst_id, IEventListener* ev);
  void RegisterComponent(const std::string& comp_id,
                         std::unique_ptr<Component::Factory> factory);

  void Start(const std::string& comp_id, const std::string& inst_id, bundle* b);
  void Resume(const std::string& inst_id);
  void Pause(const std::string& inst_id);
  bool IsResumed(const std::string& inst_id);
  void Stop(const std::string& inst_id);

  void Exit(const std::string& inst_id);
  void ExitAtIdle(const std::string& inst_id);
  void ExitAll();
  bool Bind(const std::string& inst_id, std::unique_ptr<IWindow> win);
  void Unbind(const std::string& inst_id);
  std::string GetInstanceID(int win_id);
  const IWindow* GetWindow(const std::string& inst_id);
  void AddGroup(const std::string& inst_id, int surf_id);
  void RemoveGroup(const std::string& inst_id);
  bool IsEmpty();
  void ResumeAll();
  void PauseAll();
  bool IsSingleInstance(const std::string& inst_id);

 private:
  IEventListener* GetEventListener(const std::string& inst_id);
  Ecore_Wl2_Window* GetWl2Window(const std::string& inst_id);
  void Remove(const std::string& inst_id);

 private:
  static void OnCreateCB(appcore_multiwindow_base_instance_h context,
                         void* data);
  static void OnTerminateCB(appcore_multiwindow_base_instance_h context,
                            void* data);
  static void OnPauseCB(appcore_multiwindow_base_instance_h context,
                        void* data);
  static void OnResumeCB(appcore_multiwindow_base_instance_h context,
                         void* data);

 private:
  std::map<std::string, IEventListener*> ev_map_;
  std::map<std::string, std::unique_ptr<Component::Factory>> fac_map_;
  std::map<std::string, std::unique_ptr<Component>> inst_map_;
  std::map<std::string, std::unique_ptr<IWindow>> win_map_;
  std::map<std::string, int> group_map_;
  std::list<std::string> multi_inst_list_;
};

}  // namespace internal
}  // namespace component_based

#endif  // COMPONENT_BASED_BASE_COMPONENT_MANAGER_INTERNAL_H_
