/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <app_common.h>
#include <bundle_cpp.h>

#include <algorithm>

#include "component_based/base/content_manager_internal.h"
#include "component_based/base/dlog_internal.h"

namespace component_based {
namespace internal {

ContentManager::ContentManager(std::string name) {
  auto& path = ContentManager::Path::GetInst();
  file_ = std::unique_ptr<File>(new File(path.GetPath(name)));
}

ContentManager::~ContentManager() = default;

void ContentManager::SetContent(tizen_base::Bundle content) {
  auto raw = content.ToRaw();
  uint8_t* p = reinterpret_cast<uint8_t*>(raw.first.get());

  std::vector<uint8_t> raw_data;
  std::copy(p, p + raw.second, std::back_inserter(raw_data));

  if (raw_data.empty()) {
    file_->SetRawData(raw_data);
    if (file_->IsExisting())
      file_->Delete();
  } else {
    file_->SetRawData(raw_data);
    int r = file_->Write();
    if (r < 0)
      LOGE("Failed to write raw data");
  }
}

tizen_base::Bundle ContentManager::GetContent() {
  if (file_->IsExisting()) {
    int r = file_->Read();
    if (r < 0)
      LOGE("Failed to read raw data");
  }

  auto& raw_data = file_->GetRawData();
  std::string raw(raw_data.begin(), raw_data.end());
  if (raw.empty())
    return tizen_base::Bundle();

  return tizen_base::Bundle(raw);
}

ContentManager::File::File(std::string path) : path_(std::move(path)) {
}

ContentManager::File::~File() = default;

void ContentManager::File::SetRawData(std::vector<uint8_t> raw_data) {
  raw_data_ = std::move(raw_data);
}

const std::vector<uint8_t>& ContentManager::File::GetRawData() {
  return raw_data_;
}

int ContentManager::File::Write() {
  int retry_count = 0;
  int r = TryWrite();
  while (r != 0 && retry_count < RETRY_COUNT) {
    retry_count++;
    usleep(retry_count * RETRY_SLEEP_UTIME);
    r = TryWrite();
  }

  return r;
}

int ContentManager::File::Read() {
  int retry_count = 0;
  int r = TryRead();
  while (r != 0 && retry_count < RETRY_COUNT) {
    retry_count++;
    usleep(retry_count * RETRY_SLEEP_UTIME);
    r = TryRead();
  }

  return r;
}

int ContentManager::File::Delete() {
  int retry_count = 0;
  int r = TryDelete();
  while (r != 0 && retry_count < RETRY_COUNT) {
    retry_count++;
    usleep(retry_count * RETRY_SLEEP_UTIME);
    r = TryDelete();
  }

  return r;
}

bool ContentManager::File::IsExisting() {
  if (access(path_.c_str(), F_OK) == 0)
    return true;
  return false;
}

int ContentManager::File::Open(bool readonly) {
  if (readonly)
    fd_ = open(path_.c_str(), O_RDONLY);
  else
    fd_ = open(path_.c_str(), (O_CREAT | O_WRONLY | O_TRUNC), 0644);
  if (fd_ < 0) {
    LOGE("Failed to open %s. errno(%d)", path_.c_str(), errno);
    return -1;
  }

  return 0;
}

void ContentManager::File::Close() {
  if (fd_ > 0) {
    close(fd_);
    fd_ = 0;
  }
}

int ContentManager::File::Sync() {
  int r = fdatasync(fd_);
  if (r < 0) {
    LOGE("Failed to synchronize a file. errno(%d)", errno);
    return r;
  }

  return 0;
}

bool ContentManager::File::Lock(bool readonly) {
  struct flock lock;

  lock.l_type = readonly ? F_RDLCK : F_WRLCK;
  lock.l_start = 0;
  lock.l_whence = SEEK_SET;
  lock.l_len = 0;

  if (fcntl(fd_, F_SETLK, &lock) < 0)
    return false;

  return true;
}

void ContentManager::File::Unlock() {
  struct flock lock;

  lock.l_type = F_UNLCK;
  lock.l_start = 0;
  lock.l_whence = SEEK_SET;
  lock.l_len = 0;

  fcntl(fd_, F_SETLK, &lock);
}

int ContentManager::File::TryWrite() {
  int r = Open(false);
  if (r < 0)
    return r;

  if (!Lock(false)) {
    Close();
    return -errno;
  }

  ssize_t bytes_written;
  ssize_t size = raw_data_.size();
  uint8_t* buffer = static_cast<uint8_t*>(&raw_data_[0]);

  bytes_written = write(fd_, buffer, size);
  if (bytes_written <= 0) {
    r = -errno;
    LOGE("Failed to write to a fd(%d). errno(%d)", fd_, errno);
  } else {
    r = Sync();
  }

  Unlock();
  Close();
  return r;
}

int ContentManager::File::TryRead() {
  int r = Open();
  if (r < 0)
    return r;

  if (!Lock()) {
    Close();
    return -1;
  }

  ssize_t total_size = 0;
  ssize_t bytes_read;
  uint8_t buffer[4096] = { 0, };

  raw_data_.clear();
  while ((bytes_read = read(fd_, buffer, sizeof(buffer))) > 0) {
    std::copy(buffer, buffer + bytes_read, std::back_inserter(raw_data_));
    total_size += bytes_read;
  }

  if (total_size == 0) {
    LOGE("Total size is zero");
    return -1;
  }

  Unlock();
  Close();
  return static_cast<int>(total_size);
}

int ContentManager::File::TryDelete() {
  int r = unlink(path_.c_str());
  if (r != 0) {
    LOGE("Failed to delete %s. errno(%d)", path_.c_str(), errno);
    return r;
  }
  return 0;
}

}  // namespace internal
}  // namespace component_based
