/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef COMPONENT_BASED_BASE_CORE_BASE_INTERNAL_H_
#define COMPONENT_BASED_BASE_CORE_BASE_INTERNAL_H_

#include <appcore_multiwindow_base.h>

#include <string>

namespace component_based {
namespace internal {

class CoreBase {
 public:
  class IEventListener {
   public:
    virtual void OnBaseInit(int argc, char** argv) = 0;
    virtual void OnBaseFinish() = 0;
    virtual void OnBaseRun() = 0;
    virtual void OnBaseExit() = 0;
    virtual bool OnBaseCreate() = 0;
    virtual void OnBaseTerminate() = 0;
    virtual void OnBaseReceive(aul_type type, bundle* b) = 0;
    virtual void OnBaseControl(bundle* b) = 0;

    virtual void OnWindowShow(int type, void* event) = 0;
    virtual void OnWindowHide(int type, void* event) = 0;
    virtual void OnWindowLower(int type, void* event) = 0;
    virtual void OnWindowVisibility(int type, void* event) = 0;
    virtual void OnWindowPreVisibility(int type, void* event) = 0;
    virtual void OnWindowAuxMessage(int type, void* event) = 0;
  };

  explicit CoreBase(IEventListener* ev)
    : ev_(ev), ops_(appcore_multiwindow_base_get_default_ops()) {
    ops_.base.init = OnInitCB;
    ops_.base.finish = OnFinishCB;
    ops_.base.run = OnRunCB;
    ops_.base.exit = OnExitCB;
    ops_.base.create = OnCreateCB;
    ops_.base.terminate = OnTerminateCB;
    ops_.base.control = OnControlCB;
    ops_.base.receive = OnReceiveCB;

    ops_.window.show = OnShowCB;
    ops_.window.hide = OnHideCB;
    ops_.window.lower = OnLowerCB;
    ops_.window.visibility = OnVisibilityCB;
    ops_.window.pre_visibility = OnPreVisibilityCB;
    ops_.window.aux_message = OnAuxMessageCB;
  }

  int Init(int argc, char** argv) {
    return appcore_multiwindow_base_init(ops_, argc, argv, ev_);
  }

  void Fini() {
    appcore_multiwindow_base_fini();
  }

  void Exit() {
    appcore_multiwindow_base_exit();
  }

  bool IsResumed() {
    return appcore_multiwindow_base_window_is_resumed();
  }

  virtual ~CoreBase() = default;

 private:
  static void OnInitCB(int argc, char** argv, void* data) {
    IEventListener* ev = static_cast<IEventListener*>(data);
    ev->OnBaseInit(argc, argv);
  }

  static void OnFinishCB(void) {
  }

  static void OnRunCB(void* data) {
    IEventListener* ev = static_cast<IEventListener*>(data);
    ev->OnBaseRun();
  }

  static void OnExitCB(void* data) {
    IEventListener* ev = static_cast<IEventListener*>(data);
    ev->OnBaseExit();
  }

  static int OnCreateCB(void* data) {
    appcore_multiwindow_base_on_create();
    IEventListener* ev = static_cast<IEventListener*>(data);
    if (!ev->OnBaseCreate())
      return -1;

    return 0;
  }

  static int OnTerminateCB(void* data) {
    appcore_multiwindow_base_on_terminate();
    IEventListener* ev = static_cast<IEventListener*>(data);
    ev->OnBaseTerminate();
    return 0;
  }

  static int OnReceiveCB(aul_type type, bundle* b, void* data) {
    appcore_multiwindow_base_on_receive(type, b);
    IEventListener* ev = static_cast<IEventListener*>(data);
    ev->OnBaseReceive(type, b);
    return 0;
  }

  static int OnControlCB(bundle* b, void *data) {
    appcore_multiwindow_base_on_control(b);
    IEventListener* ev = static_cast<IEventListener*>(data);
    ev->OnBaseControl(b);
    return 0;
  }

  static void OnShowCB(int type, void* event, void* data) {
    appcore_multiwindow_base_window_on_show(type, event);
    IEventListener* ev = static_cast<IEventListener*>(data);
    ev->OnWindowShow(type, event);
  }

  static void OnHideCB(int type, void* event, void* data) {
    appcore_multiwindow_base_window_on_hide(type, event);
    IEventListener* ev = static_cast<IEventListener*>(data);
    ev->OnWindowHide(type, event);
  }

  static void OnLowerCB(int type, void* event, void* data) {
    appcore_multiwindow_base_window_on_lower(type, event);
    IEventListener* ev = static_cast<IEventListener*>(data);
    ev->OnWindowLower(type, event);
  }

  static void OnVisibilityCB(int type, void* event, void* data) {
    appcore_multiwindow_base_window_on_visibility(type, event);
    IEventListener* ev = static_cast<IEventListener*>(data);
    ev->OnWindowVisibility(type, event);
  }

  static void OnPreVisibilityCB(int type, void* event, void* data) {
    appcore_multiwindow_base_window_on_pre_visibility(type, event);
    IEventListener* ev = static_cast<IEventListener*>(data);
    ev->OnWindowPreVisibility(type, event);
  }

  static void OnAuxMessageCB(int type, void* event, void* data) {
    appcore_multiwindow_base_window_on_aux_message(type, event);
    IEventListener* ev = static_cast<IEventListener*>(data);
    ev->OnWindowAuxMessage(type, event);
  }

 private:
  IEventListener* ev_;
  appcore_multiwindow_base_ops ops_;
};

}  // namespace internal
}  // namespace component_based

#endif  // COMPONENT_BASED_BASE_CORE_BASE_INTERNAL_H_
