/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <app_common.h>
#include <aul.h>
#include <aul_comp_status.h>

#include <memory>

#include "component_based/base/component_manager_internal.h"
#include "component_based/base/dlog_internal.h"
#include "component_based/base/frame_component.h"

namespace component_based {

FrameComponent::FrameComponent(std::string comp_id, std::string inst_id)
  : Component::Component(Component::Type::Frame, std::move(comp_id),
                         std::move(inst_id)) {
}

FrameComponent::~FrameComponent() = default;

FrameComponent::DisplayStatus FrameComponent::GetDisplayStatus() {
  app_display_state_e state;
  int r = app_get_display_state(&state);
  if (r != APP_ERROR_NONE) {
    LOGE("Failed to get display state. error(%d)", r);
    return FrameComponent::DisplayStatus::Unknown;
  }

  if (state == APP_DISPLAY_STATE_ON)
    return FrameComponent::DisplayStatus::On;

  return FrameComponent::DisplayStatus::Off;
}

const IWindow* FrameComponent::GetWindow() {
  auto& mgr = internal::ComponentManager::GetInst();
  return mgr.GetWindow(GetInstanceID());
}

bool FrameComponent::OnBaseCreate() {
  std::unique_ptr<IWindow> win = OnCreate();
  if (win == nullptr) {
    LOGE("OnCreate() returns nullptr");
    return false;
  }

  auto& mgr = internal::ComponentManager::GetInst();
  if (!mgr.Bind(GetInstanceID(), std::move(win))) {
    LOGE("Bind window fail");
    return false;
  }

  return true;
}

void FrameComponent::OnBaseDestroy() {
  OnDestroy();
}

void FrameComponent::OnBaseStart(AppControl control, bool restarted) {
  tizen_base::Bundle content = GetContent();
  OnBaseRestoreContent(content);
  OnStart(control, restarted);
}

void FrameComponent::OnBaseResume() {
  if (GetState() == State::Paused ||
      GetState() == State::Started ||
      GetState() == State::Stopped) {
    OnResume();
    SetState(State::Running);
    aul_comp_status_update(GetInstanceID().c_str(), COMP_STATUS_RESUMED);
  }
}

void FrameComponent::OnBasePause() {
  if (GetState() == State::Running) {
    OnPause();
    SetState(State::Paused);
    aul_comp_status_update(GetInstanceID().c_str(), COMP_STATUS_PAUSED);
  }
}

void FrameComponent::OnBaseStop() {
  OnStop();
  tizen_base::Bundle content = GetContent();
  OnBaseSaveContent(content);
  SetContent(std::move(content));
}

void FrameComponent::OnBaseRestoreContent(tizen_base::Bundle content) {
}

void FrameComponent::OnBaseSaveContent(tizen_base::Bundle& content) {
}

void FrameComponent::OnBaseAction(std::string action, AppControl app_control) {
}

void FrameComponent::OnBaseDeviceOrientationChanged(
    DeviceOrientation::Orientation orientation) {
}

void FrameComponent::OnBaseLanguageChanged(std::string language) {
}

void FrameComponent::OnBaseRegionFormatChanged(std::string region) {
}

void FrameComponent::OnBaseLowBattery(LowBattery::Status status) {
}

void FrameComponent::OnBaseLowMemory(LowMemory::Status status) {
}

void FrameComponent::OnBaseSuspendedStateChanged(SuspendedState::State state) {
}

std::unique_ptr<IWindow> FrameComponent::OnCreate() {
  return nullptr;
}

void FrameComponent::OnStart(AppControl app_control, bool restarted) {
}

void FrameComponent::OnResume() {
}

void FrameComponent::OnPause() {
}

void FrameComponent::OnStop() {
}

void FrameComponent::OnDestroy() {
}

}  // namespace component_based
