/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <dlog.h>
#include <glib.h>
#include <unistd.h>
#include <app_control_internal.h>
#include <Ecore_Wl2.h>

#include <list>

#include "component_based/base/api/component_common.h"
#include "component_based/efl_base/api/service_component.h"
#include "component_based/efl_base/api/frame_component.h"
#include "component_based/efl_base/api/component_based_app.h"

#include "component_based/common/exception.h"
#include "component_based/efl_base/application.h"
#include "component_based/efl_base/elm_window.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif
#define LOG_TAG "CAPI_COMPONENT_BASED_APPLICATION"

#ifdef EXPORT_API
#undef EXPORT_API
#endif
#define EXPORT_API __attribute__((visibility("default")))

namespace {

class ComponentClass {
 public:
  void Add(std::string comp_id,
           std::unique_ptr<component_based::Component::Factory> factory) {
    auto iter = facs_.find(comp_id);
    if (iter != facs_.end())
      LOGW("Already exists");

    facs_[comp_id] = std::move(factory);
  }

  std::map<std::string, std::unique_ptr<component_based::Component::Factory>>
  Detach() {
    return std::move(facs_);
  }

 private:
  std::map<std::string, std::unique_ptr<component_based::Component::Factory>> facs_;
};

class StubApplication : public component_based::Application {
 private:
  StubApplication() = default;
  ~StubApplication() = default;

 public:
  enum class State {
    LAUNCHED,
    CREATED,
    RUNNING,
    DYING,
  };

  static StubApplication& GetInst() {
    static StubApplication inst;
    return inst;
  }

  void SetLifecycleCallback(component_based_app_lifecycle_callback_s cb) {
    cb_ = cb;
  }

  void SetUserData(void* user_data) {
    user_data_ = user_data;
  }

  State GetState() {
    return state_;
  }

 private:
  void OnInit(int argc, char** argv) override {
    component_based::Application::OnInit(argc, argv);
  }

  void OnFini() override {
    state_ = State::DYING;
    component_based::Application::OnFini();
  }

  void OnRun() override {
    state_ = State::RUNNING;
    component_based::Application::OnRun();
  }

  void OnExit() override {
    component_based::Application::OnExit();
  }

  std::map<std::string, std::unique_ptr<component_based::Component::Factory>>
  OnCreate() override {
    std::map<std::string, std::unique_ptr<component_based::Component::Factory>> facs;
    component_class_h h = nullptr;
    if (cb_.create)
      h = cb_.create(user_data_);

    if (h == nullptr)
      return facs;

    auto cls = static_cast<ComponentClass*>(h);
    facs = std::move(cls->Detach());
    delete cls;

    state_ = State::CREATED;
    return facs;
  }

  void OnTerminate() override {
    if (cb_.terminate)
      cb_.terminate(user_data_);
  }

 private:
  State state_ = State::LAUNCHED;
  component_based_app_lifecycle_callback_s cb_;
  void* user_data_ = nullptr;
};

class StubFrameComponent : public component_based::FrameComponent {
 public:
  class Factory : public component_based::FrameComponent::Factory {
   public:
    Factory(frame_component_lifecycle_callback_s cb, void* user_data)
      : cb_(cb), user_data_(user_data) {
    }

    std::unique_ptr<component_based::Component> Create(std::string comp_id,
        std::string inst_id) override {
      return std::unique_ptr<component_based::Component>(
          new (std::nothrow) StubFrameComponent(
            std::move(comp_id), std::move(inst_id), cb_, user_data_));
    }

   private:
    frame_component_lifecycle_callback_s cb_;
    void* user_data_;
  };

  StubFrameComponent(std::string comp_id,
                     std::string inst_id,
                     frame_component_lifecycle_callback_s cb,
                     void* user_data)
    : component_based::FrameComponent(std::move(comp_id), std::move(inst_id)),
      cb_(cb),
      user_data_(user_data) {
  }

  ~StubFrameComponent() = default;

 public:
  std::unique_ptr<component_based::IWindow> OnCreate() override {
    Evas_Object* win = cb_.create(this, user_data_);
    if (win == nullptr) {
      LOGE("window is nullptr");
      return nullptr;
    }

    component_based::ElmWindow* elm_win =
        new (std::nothrow) component_based::ElmWindow(win);
    if (elm_win == nullptr) {
      LOGE("Out of memory");
      return nullptr;
    }

    return std::unique_ptr<component_based::IWindow>(
        static_cast<component_based::IWindow*>(elm_win));
  }

  void OnStart(component_based::AppControl app_control,
               bool restarted) override {
    if (cb_.start)
      cb_.start(this, app_control.GetHandle(), restarted, user_data_);
  }

  void OnResume() override {
    if (cb_.resume)
      cb_.resume(this, user_data_);
  }

  void OnPause() override {
    if (cb_.pause)
      cb_.pause(this, user_data_);
  }

  void OnStop() override {
    if (cb_.stop)
      cb_.stop(this, user_data_);
  }

  void OnDestroy() override {
    if (cb_.destroy)
      cb_.destroy(this, user_data_);
  }

  void OnBaseRestoreContent(tizen_base::Bundle content) override {
    if (cb_.restore_content)
      cb_.restore_content(this, content.GetHandle(), user_data_);
  }

  void OnBaseSaveContent(tizen_base::Bundle& content) override {
    if (cb_.save_content)
      cb_.save_content(this, content.GetHandle(), user_data_);
  }

  void OnBaseAction(std::string action,
                component_based::AppControl app_control) override {
    if (cb_.action)
      cb_.action(this, action.c_str(), app_control.GetHandle(), user_data_);
  }

  void OnBaseDeviceOrientationChanged(
      component_based::DeviceOrientation::Orientation orientation) override {
    if (cb_.device_orientation_changed) {
      cb_.device_orientation_changed(this,
          static_cast<component_device_orientation_e>(orientation),
          user_data_);
    }
  }

  void OnBaseLanguageChanged(std::string language) override {
    if (cb_.language_changed)
      cb_.language_changed(this, language.c_str(), user_data_);
  }

  void OnBaseRegionFormatChanged(std::string region) override {
    if (cb_.region_format_changed)
      cb_.region_format_changed(this, region.c_str(), user_data_);
  }

  void OnBaseLowBattery(component_based::LowBattery::Status status) override {
    if (cb_.low_battery) {
      cb_.low_battery(this,
          static_cast<component_low_battery_status_e>(status),
          user_data_);
    }
  }

  void OnBaseLowMemory(component_based::LowMemory::Status status) override {
    if (cb_.low_memory) {
      cb_.low_memory(this,
          static_cast<component_low_memory_status_e>(status),
          user_data_);
    }
  }

  void OnBaseSuspendedStateChanged(
      component_based::SuspendedState::State state) override {
    if (cb_.suspended_state_changed) {
      cb_.suspended_state_changed(this,
          static_cast<component_suspended_state_e>(state),
          user_data_);
    }
  }

 private:
  frame_component_lifecycle_callback_s cb_;
  void* user_data_;
};

class StubServiceComponent : public component_based::ServiceComponent {
 public:
  class Factory : public component_based::ServiceComponent::Factory {
   public:
    Factory(service_component_lifecycle_callback_s cb, void* user_data)
      : cb_(cb), user_data_(user_data) {
    }

    std::unique_ptr<component_based::Component> Create(std::string comp_id,
        std::string inst_id) override {
      return std::unique_ptr<component_based::Component>(
          new (std::nothrow) StubServiceComponent(std::move(comp_id),
              std::move(inst_id), cb_, user_data_));
    }

   private:
    service_component_lifecycle_callback_s cb_;
    void* user_data_;
  };

  StubServiceComponent(std::string comp_id,
                       std::string inst_id,
                       service_component_lifecycle_callback_s cb,
                       void* user_data)
    : component_based::ServiceComponent(std::move(comp_id), std::move(inst_id)),
      cb_(cb),
      user_data_(user_data) {
  }

  ~StubServiceComponent() = default;

 public:
  bool OnCreate() override {
    return cb_.create(this, user_data_);
  }

  void OnStartCommand(component_based::AppControl app_control,
                      bool restarted) override {
    if (cb_.start_command)
      cb_.start_command(this, app_control.GetHandle(), restarted, user_data_);
  }

  void OnDestroy() override {
    if (cb_.destroy)
      cb_.destroy(this, user_data_);
  }

  void OnBaseRestoreContent(tizen_base::Bundle content) override {
    if (cb_.restore_content)
      cb_.restore_content(this, content.GetHandle(), user_data_);
  }

  void OnBaseSaveContent(tizen_base::Bundle& content) override {
    if (cb_.save_content)
      cb_.save_content(this, content.GetHandle(), user_data_);
  }

  void OnBaseAction(std::string action,
                component_based::AppControl app_control) override {
    if (cb_.action)
      cb_.action(this, action.c_str(), app_control.GetHandle(), user_data_);
  }

  void OnBaseDeviceOrientationChanged(
      component_based::DeviceOrientation::Orientation orientation) override {
    if (cb_.device_orientation_changed) {
      cb_.device_orientation_changed(this,
          static_cast<component_device_orientation_e>(orientation),
          user_data_);
    }
  }

  void OnBaseLanguageChanged(std::string language) override {
    if (cb_.language_changed)
      cb_.language_changed(this, language.c_str(), user_data_);
  }

  void OnBaseRegionFormatChanged(std::string region) override {
    if (cb_.region_format_changed)
      cb_.region_format_changed(this, region.c_str(), user_data_);
  }

  void OnBaseLowBattery(component_based::LowBattery::Status status) override {
    if (cb_.low_battery) {
      cb_.low_battery(this,
          static_cast<component_low_battery_status_e>(status),
          user_data_);
    }
  }

  void OnBaseLowMemory(component_based::LowMemory::Status status) override {
    if (cb_.low_memory) {
      cb_.low_memory(this,
          static_cast<component_low_memory_status_e>(status),
          user_data_);
    }
  }

  void OnBaseSuspendedStateChanged(
      component_based::SuspendedState::State state) override {
    if (cb_.suspended_state_changed) {
      cb_.suspended_state_changed(this,
          static_cast<component_suspended_state_e>(state),
          user_data_);
    }
  }

 private:
  service_component_lifecycle_callback_s cb_;
  void* user_data_;
};

}  // namespace

extern "C" EXPORT_API int component_based_app_main(int argc, char** argv,
    component_based_app_lifecycle_callback_s *callback, void* user_data) {
  if (argc < 1 || argv == nullptr || !callback || !callback->create) {
    LOGE("Invalid parameter");
    return APP_ERROR_INVALID_PARAMETER;
  }

  auto& inst = ::StubApplication::GetInst();
  if (inst.GetState() == ::StubApplication::State::RUNNING) {
    LOGE("Already running");
    return APP_ERROR_ALREADY_RUNNING;
  }

  inst.SetLifecycleCallback(*callback);
  inst.SetUserData(user_data);

  inst.Run(argc, argv);

  return APP_ERROR_NONE;
}

extern "C" EXPORT_API int component_based_app_exit(void) {
  auto& inst = ::StubApplication::GetInst();
  inst.Exit();
  return APP_ERROR_NONE;
}

extern "C" EXPORT_API component_class_h
component_based_app_add_frame_component(
    component_class_h comp_class,
    const char* comp_id,
    frame_component_lifecycle_callback_s* callback,
    void* user_data) {
  if (comp_id == nullptr || callback == nullptr) {
    LOGE("Invalid parameter");
    set_last_result(APP_ERROR_INVALID_PARAMETER);
    return nullptr;
  }

  if (callback->create == nullptr) {
    LOGE("Invalid parameter");
    set_last_result(APP_ERROR_INVALID_PARAMETER);
    return nullptr;
  }

  ::ComponentClass* handle;
  if (comp_class == nullptr) {
    handle = new (std::nothrow) ::ComponentClass();
    if (handle == nullptr) {
      LOGE("Out of memory");
      set_last_result(APP_ERROR_OUT_OF_MEMORY);
      return nullptr;
    }
  } else {
    handle = static_cast<::ComponentClass*>(comp_class);
  }

  ::StubFrameComponent::Factory* factory =
      new (std::nothrow) ::StubFrameComponent::Factory(*callback, user_data);
  if (factory == nullptr) {
    LOGE("Out of memory");
    set_last_result(APP_ERROR_OUT_OF_MEMORY);
    return nullptr;
  }

  handle->Add(comp_id,
      std::unique_ptr<component_based::Component::Factory>(factory));

  comp_class = static_cast<component_class_h>(handle);

  set_last_result(APP_ERROR_NONE);
  return comp_class;
}

extern "C" EXPORT_API component_class_h
component_based_app_add_service_component(
    component_class_h comp_class,
    const char* comp_id,
    service_component_lifecycle_callback_s* callback,
    void* user_data) {
  if (comp_id == nullptr || callback == nullptr) {
    LOGE("Invalid parameter");
    set_last_result(APP_ERROR_INVALID_PARAMETER);
    return nullptr;
  }

  if (callback->create == nullptr) {
    LOGE("Invalid parameter");
    set_last_result(APP_ERROR_INVALID_PARAMETER);
    return nullptr;
  }

  ::ComponentClass* handle;
  if (comp_class == nullptr) {
    handle = new ::ComponentClass();
    if (handle == nullptr) {
      LOGE("Out of memory");
      set_last_result(APP_ERROR_OUT_OF_MEMORY);
      return nullptr;
    }
  } else {
    handle = static_cast<::ComponentClass*>(comp_class);
  }

  ::StubServiceComponent::Factory* factory =
      new (std::nothrow) ::StubServiceComponent::Factory(*callback, user_data);
  if (factory == nullptr) {
    LOGE("Out of memory");
    set_last_result(APP_ERROR_OUT_OF_MEMORY);
    return nullptr;
  }

  handle->Add(comp_id,
      std::unique_ptr<component_based::Component::Factory>(factory));

  comp_class = static_cast<component_class_h>(handle);

  set_last_result(APP_ERROR_NONE);
  return comp_class;
}

extern "C" EXPORT_API int frame_component_get_display_status(
    component_h context, component_display_status_e* display_status) {
  if (context == nullptr || display_status == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  component_based::Component* component =
      static_cast<component_based::Component*>(context);
  if (component->GetType() != component_based::Component::Type::Frame) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  ::StubFrameComponent* frame_component =
    dynamic_cast<::StubFrameComponent*>(component);
  if (frame_component == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  auto status = frame_component->GetDisplayStatus();
  if (status == component_based::FrameComponent::DisplayStatus::Unknown) {
    LOGE("Invalid context");
    return COMPONENT_ERROR_INVALID_CONTEXT;
  }

  *display_status = static_cast<component_display_status_e>(status);

  return COMPONENT_ERROR_NONE;
}

extern "C" EXPORT_API int frame_component_get_window(
    component_h context,
    Evas_Object** window) {
  if (context == nullptr || window == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  component_based::Component* component =
      static_cast<component_based::Component*>(context);
  if (component->GetType() != component_based::Component::Type::Frame) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  ::StubFrameComponent* frame_component =
      dynamic_cast<::StubFrameComponent*>(component);
  if (frame_component == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }
  const component_based::ElmWindow* win =
      static_cast<const component_based::ElmWindow*>(
          frame_component->GetWindow());
  component_based::ElmWindow* elm_win =
      const_cast<component_based::ElmWindow*>(win);
  *window = static_cast<Evas_Object*>(elm_win->GetRaw());

  return COMPONENT_ERROR_NONE;
}

extern "C" EXPORT_API int frame_component_get_resource_id(Evas_Object* win,
    int* res_id) {
  if (win == nullptr || res_id == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }
  Evas* evas = evas_object_evas_get(win);
  if (evas == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  Ecore_Evas* ecore_evas = ecore_evas_ecore_evas_get(evas);
  if (ecore_evas == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  Ecore_Wl2_Window* wl_win = ecore_evas_wayland2_window_get(ecore_evas);
  if (wl_win == nullptr) {
    LOGE("Invalid parameter");
    return COMPONENT_ERROR_INVALID_PARAMETER;
  }

  *res_id = ecore_wl2_window_id_get(wl_win);
  return COMPONENT_ERROR_NONE;
}