// Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
// Use of this source code is governed by a apache 2.0 license that can be
// found in the LICENSE file.

#include <gmock/gmock.h>
#include <gtest/gtest.h>
#include <stdlib.h>

#include <bundle_cpp.h>

#include "component_based/app_control/component_based_app_control.h"
#include "component_based/base/component.h"
#include "component_based/base/frame_component.h"
#include "component_based/base/component_manager_internal.h"
#include "component_based/base/window_interface.h"

#include "unit_tests/mock/mock_appcore_multiwindow_base.h"

namespace {

class ComponentManagerTest : public ::testing::Test {
 protected:
  void SetUp() override {}
  void TearDown() override {}
};

class SimpleComponentManagerListener : public component_based::internal::ComponentManager::IEventListener {
  bool OnCreate() override {
    return true;
  };
  void OnDestroy() override {};
  void OnStart(component_based::AppControl app_control, bool restarted) override {};
  void OnPause() override {};
  void OnResume() override {};
  void OnStop() override {};
};

class SimpleFrameComp : public component_based::FrameComponent {
 public:
  class Factory : public component_based::Component::Factory {
   public:
    std::unique_ptr<component_based::Component> Create(std::string comp_id,
                                            std::string inst_id) override {
      return std::unique_ptr<component_based::Component>(
          new SimpleFrameComp(std::move(comp_id), std::move(inst_id)));
    }
  };

  SimpleFrameComp(std::string comp_id, std::string inst_id)
    : FrameComponent(std::move(comp_id), std::move(inst_id)) {
  }

  void OnStart(component_based::AppControl app_control, bool restarted) override {
  }

  void OnResume() override {
  }

  void OnPause() override {
  }

  void OnStop() override {
  }

  void OnDestroy() override {
  }
};

class SimpleWindow : public component_based::IWindow {
 public:
  int GetResID() const override {
    return 0;
  };

  void* GetRaw() override {
    void* temp = nullptr;
    return temp;
  };
};

appcore_multiwindow_base_instance_h __fake_appcore_multiwindow_base_instance_run(
    const char* class_id, const char* id, void* extra)
{
  return nullptr;
}

appcore_multiwindow_base_instance_h __fake_appcore_multiwindow_base_instance_find(const char* id)
{
  typedef struct _appcore_multiwindow_base_instance {
    unsigned int window_id;
    char *id;
    void *extra;
    appcore_multiwindow_base_class *shell;
    bool is_resumed;
  } appcore_multiwindow_base_instance;

  appcore_multiwindow_base_instance *s =
    (appcore_multiwindow_base_instance*)malloc(sizeof(appcore_multiwindow_base_instance));
  return s;
}

bool __fake_appcore_multiwindow_base_instance_is_resumed(appcore_multiwindow_base_instance_h h)
{
  free(h);
  return 0;
}

void __fake_appcore_multiwindow_base_instance_exit(appcore_multiwindow_base_instance_h h)
{
  free(h);
}

void __fake_appcore_multiwindow_base_window_unbind(appcore_multiwindow_base_instance_h h)
{
  free(h);
}

TEST_F(ComponentManagerTest, TestIsEmpty) {
  auto &cm = component_based::internal::ComponentManager::GetInst();
  EXPECT_TRUE(cm.IsEmpty());
}

TEST_F(ComponentManagerTest, TestRegisterComponent) {
  auto &cm = component_based::internal::ComponentManager::GetInst();

  cm.RegisterComponent("comp_id",
      std::unique_ptr<component_based::Component::Factory>(
          new SimpleFrameComp::Factory()));
}

TEST_F(ComponentManagerTest, TestAddEventListener) {
  auto &cm = component_based::internal::ComponentManager::GetInst();
  std::unique_ptr<SimpleComponentManagerListener> ev(
      new SimpleComponentManagerListener());

  cm.AddEventListener(std::string("inst_id"), ev.get());
}

TEST_F(ComponentManagerTest, TestComponentManagerLifeCycle) {
  appcore_multiwindow_base_instance_run_fake.custom_fake =
    __fake_appcore_multiwindow_base_instance_run;
  appcore_multiwindow_base_instance_find_fake.custom_fake =
    __fake_appcore_multiwindow_base_instance_find;
  appcore_multiwindow_base_instance_is_resumed_fake.custom_fake =
    __fake_appcore_multiwindow_base_instance_is_resumed;
  appcore_multiwindow_base_instance_exit_fake.custom_fake =
    __fake_appcore_multiwindow_base_instance_exit;
  appcore_multiwindow_base_window_unbind_fake.custom_fake =
    __fake_appcore_multiwindow_base_window_unbind;

  auto &cm = component_based::internal::ComponentManager::GetInst();
  std::string inst_id("inst_id2");
  std::string comp_id("comp_id2");

  cm.RegisterComponent(comp_id,
      std::unique_ptr<component_based::Component::Factory>(
          new SimpleFrameComp::Factory()));

  std::unique_ptr<SimpleComponentManagerListener> ev(
      new SimpleComponentManagerListener());
  cm.AddEventListener(inst_id, ev.get());

  bundle *b = bundle_create();
  cm.Start(comp_id, inst_id, b);
  bundle_free(b);
  cm.Pause(inst_id);
  cm.PauseAll();
  cm.Resume(inst_id);
  cm.ResumeAll();
  EXPECT_FALSE(cm.IsResumed(inst_id));
  cm.Stop(inst_id);
  cm.Exit(inst_id);
  cm.ExitAll();
}

TEST_F(ComponentManagerTest, TestComponentManagerBindUnbind) {
  auto &cm = component_based::internal::ComponentManager::GetInst();
  std::string inst_id("inst_id");
  std::unique_ptr<SimpleWindow> win(new SimpleWindow());

  cm.Bind(inst_id, std::move(win));
  cm.GetWindow(inst_id);
  cm.Unbind(inst_id);
}

TEST_F(ComponentManagerTest, TestComponentManagerAddRemoveGroup) {
  auto &cm = component_based::internal::ComponentManager::GetInst();
  std::string inst_id("inst_id");

  cm.AddGroup(inst_id, 1);
  cm.RemoveGroup(inst_id);
}

}  // namespace
