/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdlib.h>
#include <gtest/gtest.h>

#include <iostream>
#include <memory>

#include "component_based/uri/component_based_uri.h"

// The source of the following example is "https://en.wikipedia.org/wiki/Uniform_Resource_Identifier"
static const char EXAM_URI1[] = "https://john.doe@www.example.com:123/forum/questions/?tag=networking&order=newest#top";
static const char EXAM_URI2[] = "ldap://[2001:db8::7]/c=GB?objectClass?one";
static const char EXAM_URI3[] = "mailto:John.Doe@example.com";
static const char EXAM_URI4[] = "news:comp.infosystems.www.servers.unix";
static const char EXAM_URI5[] = "tel:+1-816-555-1212";
static const char EXAM_URI6[] = "telnet://192.0.2.16:80/";
static const char EXAM_URI7[] = "urn:oasis:names:specification:docbook:dtd:xml:4.1.2";

static const char ENCODED_URI1[] = "https:%2F%2Fjohn.doe%40www.example.com%3A123%2Fforum%2Fquestions%2F?tag=networking&order=newest#top";
static const char ENCODED_URI2[] = "ldap:%2F%2F%5B2001%3Adb8%3A%3A7%5D%2Fc%3DGB?objectClass?one";
static const char ENCODED_URI3[] = "mailto:John.Doe%40example.com";
static const char ENCODED_URI4[] = "news:comp.infosystems.www.servers.unix";
static const char ENCODED_URI5[] = "tel:%2B1-816-555-1212";
static const char ENCODED_URI6[] = "telnet:%2F%2F192.0.2.16%3A80%2F";
static const char ENCODED_URI7[] = "urn:oasis%3Anames%3Aspecification%3Adocbook%3Adtd%3Axml%3A4.1.2";

extern int gargc;
extern char** gargv;

class UriTest : public testing::Test {
 public:
  virtual ~UriTest() {}

  virtual void SetUp() {
    std::string exam_uri[] = {
      ENCODED_URI1,
      ENCODED_URI2,
      ENCODED_URI3,
      ENCODED_URI4,
      ENCODED_URI5,
      ENCODED_URI6,
      ENCODED_URI7,
    };

    for (int i = 0; i < 7; i++) {
      uri_list_[i] = std::unique_ptr<component_based::Uri>(
          new component_based::Uri(exam_uri[i]));
    }
  }

  virtual void TearDown() {}

  std::unique_ptr<component_based::Uri> uri_list_[7];
};

/*
 * @testcase Uri_Constructor
 * @description Creates a Uri object.
 * @apicovered component_based::Uri
 */
TEST_F(UriTest, Uri_Constructor) {
  std::unique_ptr<component_based::Uri> uri =
      std::unique_ptr<component_based::Uri>(
          new component_based::Uri(ENCODED_URI1));
  EXPECT_NE(uri, nullptr);
}

/*
 * @testcase Uri_GetScheme
 * @description Gets a scheme component from a URI object.
 * @apicovered component_based::Uri::GetScheme
 */
TEST_F(UriTest, Uri_GetScheme) {
  std::string exam_scheme[] = {
    "https",
    "ldap",
    "mailto",
    "news",
    "tel",
    "telnet",
    "urn",
  };

  for (int i = 0; i < 7; i++) {
    EXPECT_EQ(uri_list_[i]->GetScheme(), exam_scheme[i]);
  }
}

/*
 * @testcase Uri_GetAuthority
 * @description Gets an authority component from a URI object.
 * @apicovered component_based::Uri::GetAuthority
 */
TEST_F(UriTest, Uri_GetAuthority) {
  std::string exam_auth[] = {
    "//john.doe@www.example.com:123",
    "//[2001:db8::7]",
    "",
    "",
    "",
    "//192.0.2.16:80",
    "",
  };

  for (int i = 0; i < 7; i++) {
    EXPECT_EQ(uri_list_[i]->GetAuthority(), exam_auth[i]);
  }
}

/*
 * @testcase Uri_GetPath
 * @description Gets a path component from a URI object.
 * @apicovered component_based::Uri::GetPath
 */
TEST_F(UriTest, Uri_GetPath) {
  std::string exam_path[] = {
    "/forum/questions/",
    "/c=GB",
    "John.Doe@example.com",
    "comp.infosystems.www.servers.unix",
    "+1-816-555-1212",
    "/",
    "oasis:names:specification:docbook:dtd:xml:4.1.2",
  };

  for (int i = 0; i < 7; i++) {
    EXPECT_EQ(uri_list_[i]->GetPath(), exam_path[i]);
  }
}

/*
 * @testcase Uri_GetFragment
 * @description Gets a fragment component from a URI object.
 * @apicovered component_based::Uri::GetFragment
 */
TEST_F(UriTest, Uri_GetFragment) {
  std::string exam_fragment[] = {
    "top",
    "",
    "",
    "",
    "",
    "",
    "",
  };

  for (int i = 0; i < 7; i++) {
    EXPECT_EQ(uri_list_[i]->GetFragment(), exam_fragment[i]);
  }
}

/*
 * @testcase Uri_GetHost
 * @description Gets a host subcomponent from a URI object.
 * @apicovered component_based::Uri::GetHost
 */
TEST_F(UriTest, Uri_GetHost) {
  std::string exam_host[] = {
    "www.example.com",
    "[2001:db8::7]",
    "",
    "",
    "",
    "192.0.2.16",
    "",
  };

  for (int i = 0; i < 7; i++) {
    EXPECT_EQ(uri_list_[i]->GetHost(), exam_host[i]);
  }
}

/*
 * @testcase Uri_GetPort
 * @description Gets a port subcomponent from a URI object.
 * @apicovered component_based::Uri::GetPort
 */
TEST_F(UriTest, Uri_GetPort) {
  std::string exam_port[] = {
    "123",
    "",
    "",
    "",
    "",
    "80",
    "",
  };

  for (int i = 0; i < 7; i++) {
    EXPECT_EQ(uri_list_[i]->GetPort(), exam_port[i]);
  }
}

/*
 * @testcase Uri_GetUser
 * @description Gets a user subcomponent from a URI object.
 * @apicovered component_based::Uri::GetUser
 */
TEST_F(UriTest, Uri_GetUser) {
  std::string exam_user[] = {
    "john.doe",
    "",
    "",
    "",
    "",
    "",
    "",
  };

  for (int i = 0; i < 7; i++) {
    EXPECT_EQ(uri_list_[i]->GetUser(), exam_user[i]);
  }
}

/*
 * @testcase Uri_GetQuery
 * @description Gets a query component from a URI object.
 * @apicovered component_based::Uri::GetQuery
 */
TEST_F(UriTest, Uri_GetQuery) {
  int match_count = 0;
  auto& query = uri_list_[0]->GetQuery();
  for (auto& q : query) {
    if (q.GetKey() == "tag" &&
        q.GetValue() == "networking") {
      match_count++;
    } else if (q.GetKey() == "order" &&
               q.GetValue() == "newest") {
      match_count++;
    }
  }

  EXPECT_EQ(match_count, 2);
}

/*
 * @testcase Uri_Encode
 * @description Encodes a URI object to string.
 * @apicovered component_base::Uri::Encode
 */
TEST_F(UriTest, Uri_Encode) {
  std::unique_ptr<component_based::Uri> uri =
      std::unique_ptr<component_based::Uri>(
          new component_based::Uri(EXAM_URI1));
  EXPECT_NE(uri, nullptr);
  EXPECT_EQ(uri->Encode(), ENCODED_URI1);
}

/*
 * @testcase Uri_Builder
 * @description Builds a URI builder object to create a URI object.
 * @apicovered component_based::Uri::Builder,
 *             component_based::Uri::Builder::SetScheme,
 *             component_based::Uri::Builder::SetAuthority,
 *             component_based::Uri::Builder::SetPath,
 *             component_based::Uri::Builder::AddQuery,
 *             component_based::Uri::Builder::SetFragment,
 *             component_based::Uri::Builder::Build
 */
TEST_F(UriTest, Uri_Builder) {
  std::unique_ptr<component_based::Uri::Builder> builder(
      new component_based::Uri::Builder());
  EXPECT_NE(builder, nullptr);
  builder->SetScheme("https");
  builder->SetAuthority("//john.doe@www.example.com:123");
  builder->SetPath("/forum");
  builder->AddPath("questions/");
  builder->AddQuery("tag", "networking");
  builder->AddQuery("order", "newest");
  builder->SetFragment("top");

  component_based::Uri uri = builder->Build();
  EXPECT_EQ(uri.Encode(), ENCODED_URI1);
}
